@@ -1,5 +1,7 @@
 package Test::CheckManifest;
 
+# ABSTRACT: Check if your Manifest matches your distro
+
 use strict;
 use warnings;
 
@@ -10,7 +12,7 @@ use File::Basename;
 use Test::Builder;
 use File::Find;
 
-our $VERSION = '1.0';
+our $VERSION = '1.25_01';
 
 my $test      = Test::Builder->new();
 my $test_bool = 1;
@@ -43,10 +45,28 @@ sub ok_manifest{
     
     $msg = $hashref unless $is_hashref;
     
-    my $bool     = 1;
-    my $dir      =
-    my $home     = Cwd::realpath(dirname(File::Spec->rel2abs($0)) . '/..');    
-    my $manifest = Cwd::realpath($home . '/MANIFEST');
+    my $bool = 1;
+    my $home = Cwd::realpath( dirname ( File::Spec->rel2abs($0) ) );
+    my $manifest;
+
+    my $counter = 0;
+    while ( 1 ) {
+        my $tmp_home = Cwd::realpath( $home . '/..' );
+
+        last if $tmp_home eq $home || $counter++ == 20;
+        $home = $tmp_home;
+
+        my $manifest_path = File::Spec->catfile( $home . '/MANIFEST' );
+        last if -e $manifest_path;
+    }
+
+    eval { $manifest = Cwd::realpath( $home . '/MANIFEST' ); 1; };
+    if ( !$manifest ) {
+        $test->BAILOUT( 'Cannot find a MANIFEST. Please check!' );
+    }
+    
+    my $skip;
+    eval { $skip     = Cwd::realpath( $home . '/MANIFEST.SKIP' ); 1; };
     
     my @missing_files = ();
     my @files_plus    = ();
@@ -59,9 +79,9 @@ sub ok_manifest{
                                'and' :
                                'or'; 
                    
-    push @$arref, @{$hashref->{exclude}} if($is_hashref and
-                                            exists $hashref->{exclude} and 
-                                            ref($hashref->{exclude}) eq 'ARRAY');
+    push @$arref, @{$hashref->{exclude}} 
+        if $is_hashref and exists $hashref->{exclude} and 
+            ref($hashref->{exclude}) eq 'ARRAY';
     
     for(@$arref){
         croak 'path in excluded array must be "absolute"' unless m!^/!;
@@ -72,12 +92,17 @@ sub ok_manifest{
     
     @$arref = grep { defined }@$arref;
     
-    unless( open(my $fh,'<',$manifest) ){
+    unless( open my $fh, '<', $manifest ){
         $bool = 0;
         $msg  = "can't open $manifest";
     }
     else{
-        my @files = grep{!/^\s*$/ and !/^META\.yml/}<$fh>;
+        { # extra block to use "last"
+        
+        my $files_in_skip = _read_skip( $skip, \$msg, \$bool );
+        last unless $files_in_skip;
+            
+        my @files = _read_file( $fh );
         close $fh;
     
         chomp @files;
@@ -97,18 +122,31 @@ sub ok_manifest{
         @files_hash{@files} = ();
     
         find({no_chdir => 1,
-          wanted   => sub{ my $file         = $File::Find::name;
-                           my $is_excluded  = _is_excluded($file,$arref,$filter,$comb);
-                           push(@dir_files,Cwd::realpath($file)) if -f $file and !$is_excluded;
-                           $excluded{$file} = 1 if -f $file and $is_excluded}},$home);
+            wanted   => sub{
+                my $file         = $File::Find::name;
+                my $is_excluded  = _is_excluded(
+                    $file,
+                    $arref,
+                    $filter,
+                    $comb,
+                    $files_in_skip,
+                    $home,
+                );
+                
+                push(@dir_files,Cwd::realpath($file)) if -f $file and !$is_excluded;
+                
+                $excluded{$file} = 1 if -f $file and $is_excluded
+            }
+        },$home);
 
-    
+        #use Data::Dumper;
         #print STDERR ">>",++$counter,":",Dumper(\@files,\@dir_files);
-        CHECK: for my $file(@dir_files){
+        SFILE:
+        for my $file(@dir_files){
             for my $check(@files){
                 if($file eq $check){
                     delete $files_hash{$check};
-                    next CHECK;
+                    next SFILE;
                 }
             }
             push(@missing_files,$file);
@@ -116,8 +154,10 @@ sub ok_manifest{
         }
     
         delete $files_hash{$_} for keys %excluded;
-        @files_plus = keys %files_hash;
-        $bool = 0 if scalar @files_plus > 0;    
+        @files_plus = sort keys %files_hash;
+        $bool = 0 if scalar @files_plus > 0;
+        
+        } # close extra block
     }
     
     my $diag = 'The following files are not named in the MANIFEST file: '.
@@ -130,6 +170,32 @@ sub ok_manifest{
     $test->diag($plus) if scalar @files_plus    >= 1 and $test_bool == 1;
 }
 
+sub _read_file {
+    my ($fh) = @_;
+    
+    my @files;
+    while( my $fh_line = <$fh> ){
+        chomp $fh_line;
+        
+        next if $fh_line =~ m{ \A \s* \# }x;
+        
+        my ($file);
+        
+        if ( ($file) = $fh_line =~ /^'(\\[\\']|.+)+'\s*(.*)/) {
+            $file =~ s/\\([\\'])/$1/g;
+        }
+        else {
+            ($file) = $fh_line =~ /^(\S+)\s*(.*)/;
+        }
+
+        next unless $file;
+        
+        push @files, $file;
+    }
+    
+    return @files;
+}
+
 sub _not_ok_manifest{
     $test_bool = 0;
     ok_manifest(@_);
@@ -137,11 +203,19 @@ sub _not_ok_manifest{
 }
 
 sub _is_excluded{
-    my ($file,$dirref,$filter,$bool) = @_;
+    my ($file,$dirref,$filter,$bool,$files_in_skip,$home) = @_;
     my @excluded_files = qw(pm_to_blib Makefile META.yml Build pod2htmd.tmp
-                            pod2htmi.tmp Build.bat .cvsignore);
+                            pod2htmi.tmp Build.bat .cvsignore MYMETA.json);
+
+    if ( $files_in_skip and 'ARRAY' eq ref $files_in_skip ) {
+        (my $local_file = $file) =~ s{\Q$home\E/?}{};
+        for my $rx ( @{$files_in_skip} ) {
+            my $regex = qr/$rx/;
+            return 1 if $local_file =~ $regex;
+        }
+    }
         
-    my @matches = grep{$file =~ /$_$/}@excluded_files;
+    my @matches = grep{ $file =~ /$_$/ }@excluded_files;
     
     if($bool eq 'or'){
         push @matches, $file if grep{ref($_) and ref($_) eq 'Regexp' and $file =~ /$_/}@$filter;
@@ -157,13 +231,37 @@ sub _is_excluded{
     return scalar @matches;
 }
 
+sub _read_skip {
+    my ($skip, $msg, $bool) = @_;
+
+    return [] unless $skip and -e $skip;
+    
+    my @files;
+    if( -e $skip and not open my $skip_fh, '<', $skip ) {
+        $$bool = 0;
+        $$msg  = "can't open $skip";
+        return;
+    }
+    else {
+        @files = _read_file( $skip_fh );
+    }
+
+    return \@files;
+}
+
 1;
-__END__
+
+
+=pod
 
 =head1 NAME
 
 Test::CheckManifest - Check if your Manifest matches your distro
 
+=head1 VERSION
+
+version 1.25_01
+
 =head1 SYNOPSIS
 
   use Test::CheckManifest;
@@ -179,7 +277,7 @@ There is only one method exported: C<ok_manifest>
 
 =head1 METHODS
 
-=head2 ok_manifest   [{exlude => $arref}][$msg]
+=head2 ok_manifest   [{exclude => $arref}][$msg]
 
 checks whether the Manifest file matches the distro or not. To match a distro
 the Manifest has to name all files that come along with the distribution.
@@ -240,6 +338,14 @@ These files not:
 
 =back
 
+=head1 EXCLUDING FILES
+
+Beside C<filter> and C<exclude> there is another way to exclude files:
+C<MANIFEST.SKIP>. This is a file with filenames that should be excluded:
+
+  t/my_very_own.t
+  file_to.skip
+
 =head1 ACKNOWLEDGEMENT
 
 Great thanks to Christopher H. Laco, who did a lot of testing stuff for me and
@@ -247,15 +353,18 @@ he reported some bugs to RT.
 
 =head1 AUTHOR
 
-Renee Baecker, E<lt>module@renee-baecker.deE<gt>
+Renee Baecker <module@renee-baecker.de>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2006 by Renee Baecker
+This software is Copyright (c) 2012 by Renee Baecker.
 
-This library is free software; you can redistribute it and/or modify
-it under the same terms as Perl itself, either Perl version 5.8.6 or,
-at your option, any later version of Perl 5 you may have available.
+This is free software, licensed under:
 
+  The Artistic License 2.0 (GPL Compatible)
 
 =cut
+
+
+__END__
+
